"""
This is the utils module related to the main
editor.
"""
from yta_validation import PythonValidator
from yta_programming.decorators.requires_dependency import requires_dependency
from typing import Union


@requires_dependency('numpy', 'yta_editor_utils', 'numpy')
@requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
def frame_to_texture(
    frame: Union['VideoFrame', 'np.ndarray'],
    context: 'moderngl.Context',
    numpy_format: str = 'rgb24'
):
    """
    *Optional dependency `numpy` (imported as `numpy`) required*

    *Optional dependency `moderngl` (imported as `moderngl`) required*
    
    Transform the given 'frame' to an opengl
    texture. The frame can be a VideoFrame
    instance (from pyav library) or a numpy
    array.

    (!) This method is useful to transform a
    frame into a texture quick and for a single
    use, but we have the GPUTextureHandler class
    to handle it in an specific contexto to 
    optimize the performance and avoid creating
    textures but rewriting on them.
    """
    import numpy as np
    import moderngl

    # To numpy RGB inverted for opengl
    frame: np.ndarray = (
        np.flipud(frame.to_ndarray(format = numpy_format))
        if PythonValidator.is_instance_of(frame, 'VideoFrame') else
        np.flipud(frame)
    )

    # Sometimes we have 'float32' values but we need to
    # force 'uint8' to be able to work with
    if frame.dtype != np.uint8:
        frame = np.clip(frame, 0, 255).astype(np.uint8)

    return context.texture(
        size = (frame.shape[1], frame.shape[0]),
        components = frame.shape[2],
        data = frame.tobytes()
    )

# TODO: I should make different methods to
# obtain a VideoFrame or a numpy array frame
@requires_dependency('numpy', 'yta_editor_utils', 'numpy')
@requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
def texture_to_frame(
    texture: 'moderngl.Texture',
    do_include_alpha: bool = True
) -> 'np.ndarray':
    """
    *Optional dependency `numpy` (imported as `numpy`) required*

    *Optional dependency `moderngl` (imported as `moderngl`) required*

    Transform an opengl texture into numpy array.

    The `do_include_alpha` will include the
    alpha channel if True.
    """
    import numpy as np
    import moderngl

    data = texture.read(alignment = 1)
    # Read 4 channels always (RGBA8)
    frame = np.frombuffer(data, dtype = np.uint8).reshape((texture.size[1], texture.size[0], 4))
    # TODO: Do this with a utils
    frame = (
        frame
        if do_include_alpha else
        # Discard alpha channel if not needed
        frame[:, :, :3]
    )
    # Opengl gives it with the 'y' inverted
    frame = np.flipud(frame)
    # TODO: This can be returned as a numpy frame

    # This is if we need an 'av' VideoFrame (to
    # export through the demuxer, for example)
    # TODO: I avoid this by now because we don't
    # want to import the pyav library, so this
    # below has to be done with the numpy array
    # received...
    # frame = av.VideoFrame.from_ndarray(frame, format = 'rgba')
    # # TODO: Make this customizable
    # frame = frame.reformat(format = 'yuv420p')

    return frame

# TODO: The code of this one is similar to the
# 'texture_to_frame', so we should keep only one
# (maybe this one)
@requires_dependency('numpy', 'yta_editor_utils', 'numpy')
@requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
def texture_to_numpy(
    texture: 'moderngl.Texture',
    do_include_alpha: bool = True
) -> 'np.ndarray':
    """
    *Optional dependency `numpy` (imported as `numpy`) required*

    *Optional dependency `moderngl` (imported as `moderngl`) required*

    Transform the provided OpenGL `texture` into a numpy
    array to be able to validate the values that are
    contained in the texture.

    The alpha layer will be removed if the parameter
    `do_include_alpha` is set as False.
    """
    import numpy as np
    import moderngl

    data = texture.read()

    # OpenGL uses only f4 and u1 for the textures so
    # we don't need a mapping
    dtype = (
        np.float32
        if texture.dtype == 'f4' else
        np.uint8
    )

    frame = np.frombuffer(
        buffer = data,
        dtype = dtype
    ).reshape(
        texture.height,
        texture.width,
        texture.components
    )

    # Discard alpha channel if not needed
    frame = (
        frame
        if do_include_alpha else
        frame[:, :, :3]
    )

    # Opengl gives it with the 'y' inverted
    frame = np.flipud(frame)

    return frame

# TODO: I think I have this methods in other library
# maybe yta-image-utils (?)
@requires_dependency('cv2', 'yta_editor_utils', 'opencv-python')
def scale_numpy(
    input: 'np.ndarray',
    size: tuple[int, int]
) -> 'np.ndarray':
    """
    *Optional dependency `opencv-python` required*

    Resize the 'input' numpy array to the provided 'size'
    if needed, using a rescaling method with 'opencv-python'
    (cv2).

    The 'size' provided must be (width, height).
    """
    import cv2

    return cv2.resize(input, size, interpolation = cv2.INTER_LINEAR)

@requires_dependency('PIL', 'yta_editor_utils', 'pillow')
@requires_dependency('numpy', 'yta_editor_utils', 'numpy')
def scale_numpy_pillow(
    input: 'np.ndarray',
    size: tuple[int, int]
) -> 'np.ndarray':
    """
    *Optional dependency `pillow` (imported as `PIL`) required*

    *Optional dependency `numpy` (imported as `numpy`) required*

    Resize the 'input' numpy array to the provided 'size'
    if needed, using a rescaling method with 'pillow'
    (PIL).

    The 'size' provided must be (width, height).
    """
    from PIL import Image
    import numpy as np

    return np.array(Image.fromarray(input).resize(size, Image.BILINEAR))

@requires_dependency('PIL', 'yta_editor_utils', 'pillow')
@requires_dependency('numpy', 'yta_editor_utils', 'numpy')
def read_image_as_numpy(
    path: str,
    do_read_as_rgba: bool = True,
    size: Union[tuple[int, int], None] = None
) -> 'np.ndarray':
    """
    *Optional dependency `pillow` (imported as `PIL`) required*

    *Optional dependency `numpy` (imported as `numpy`) required*

    Read an image from a file and transform it into a
    numpy array. It will force the 'size' if provided,
    or leave the original one if it is None.
    """
    from PIL import Image
    import numpy as np

    mode = (
        'RGBA'
        if do_read_as_rgba else
        'RGB'
    )

    with Image.open(path) as img:
        img = img.convert(mode)
        np_img = np.array(img, dtype = np.uint8)

    return (
        scale_numpy_pillow(
            input = np_img,
            size = size
        ) if size is not None else
        np_img
    )

@requires_dependency('PIL', 'yta_editor_utils', 'pillow')
def numpy_to_file(
    input: 'np.ndarray',
    output_filename: str
) -> str:
    """
    *Optional dependency `pillow` (imported as `PIL`) required*

    Export the provided 'array' numpy array as a file
    with the given 'output_filename' name.
    """
    from PIL import Image

    Image.fromarray(input).save(output_filename)

    return output_filename

@requires_dependency('PIL', 'yta_editor_utils', 'pillow')
def texture_to_file(
    texture: 'moderngl.Texture',
    output_filename: str
) -> str:
    """
    *Optional dependency `pillow` (imported as `PIL`) required*

    Export the provided OpenGL texture 'texture' as a
    file with the given 'output_filename' name.
    """
    return numpy_to_file(
        input = texture_to_frame(texture),
        output_filename = output_filename
    )


"""
A selection mask is a numpy array or a texture that
indicates where we should apply the changes from
the processed input in the original input, and also
how much we have todo.

A selection mask that is fulfilled with all ones will
make the output be the processed input, but one with
all zeros will keep the original input as not modified.

(!) If a moderngl texture is detected as the input this
will trigger the detector to raise an Exception if the
library is not installed and it cannot be processed.

TODO: I think this module is very similar to the 
`yta-numpy` library we have.
"""
# TODO: Should we move this module from here to
# another different one
from yta_validation import PythonValidator
from yta_programming.decorators.requires_dependency import requires_dependency
from typing import Union


"""
Here we handle the dtypes as strings that
are accepted by opengl ('f4' for 'float32', 'u1'
for 'uint8') where f means float and 4 means
4 * 8 = 32
"""
class _TextureSelectionMaskGenerator:
    """
    *For internal use only*

    Class to be used as a shortcut within the general
    selection mask generator, to simplify the way we
    create selection masks as moderngl textures.
    """

    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_full_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a full mask (all values are 1.0) for the
        that fits the dimensions and properties of
        the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_full_mask_for_input(
                input = input,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )

    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_full_mask(
        width: int,
        height: int,
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a full mask (all values are 1.0) of the
        provided `width` and `height` given, and the
        `dtype` provided, ready to be used in an
        OpenGL shader.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_full_mask(
                width = width,
                height = height,
                number_of_channels = 4,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_half_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a half mask (all values are 0.5) for the
        that fits the dimensions and properties of
        the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_half_mask_for_input(
                input = input,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_half_mask(
        width: int,
        height: int,
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a half mask (all values are 0.5) of the
        provided `width` and `height` given, and the
        `dtype` provided, ready to be used in an
        OpenGL shader.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_half_mask(
                width = width,
                height = height,
                number_of_channels = 4,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_empty_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get an empty mask (all values are 0.0) for the
        that fits the dimensions and properties of
        the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_empty_mask_for_input(
                input = input,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_empty_mask(
        width: int,
        height: int,
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get an empty mask (all values are 0.0) of the
        provided `width` and `height` given, and the
        `dtype` provided, ready to be used in an
        OpenGL shader.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_empty_mask(
                width = width,
                height = height,
                number_of_channels = 4,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_random_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a random mask (all values are random values
        in the range [0.0, 1.0]) that fits the
        dimensions and properties of the `input` provided,
        returning it with the given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_random_mask_for_input(
                input = input,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )

    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_random_mask(
        width: int,
        height: int,
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a random mask (all values are random values
        in the range [0.0, 1.0]) of the provided `width`
        and `height` given, and the `dtype` provided,
        ready to be used in an OpenGL shader.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_random_mask(
                width = width,
                height = height,
                number_of_channels = 4,
                dtype = dtype
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_custom_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
        value: float = 0.75,
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a custom mask (all values are the provided
        `value`, that must be a value in the [0.0, 1.0]
        range) that fits the dimensions and properties
        of the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_custom_mask_for_input(
                input = input,
                dtype = dtype,
                value = value
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def get_custom_mask(
        width: int,
        height: int,
        dtype: str = 'f4',
        value: float = 0.75,
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        Get a custom mask (all values are the provided
        `value`, that must be a value in the [0.0, 1.0]
        range) of the provided `width` and `height` and
        the `dtype` provided.
        """
        return _NumpySelectionMaskGenerator._numpy_mask_to_texture(
            mask = _NumpySelectionMaskGenerator.get_custom_mask(
                width = width,
                height = height,
                number_of_channels = 4,
                dtype = dtype,
                value = value
            ),
            dtype = dtype,
            opengl_context = opengl_context
        )

class _NumpySelectionMaskGenerator:
    """
    *For internal use only*

    Class to be used as a shortcut within the general
    selection mask generator, to simplify the way we
    create seletion masks as numpy arrays.
    """

    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_full_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
    ) -> 'np.ndarray':
        """
        Get a full mask (all values are 1.0) that fits
        the dimensions and properties of the `input`
        provided, returning it with the given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        width, height, number_of_channels = _get_width_height_and_number_of_channels_from_input(input)

        return _NumpySelectionMaskGenerator.get_full_mask(
            width = width,
            height = height,
            number_of_channels = number_of_channels,
            dtype = dtype
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_full_mask(
        width: int,
        height: int,
        number_of_channels: int,
        dtype: str = 'f4'
    ) -> 'np.ndarray':
        """
        Get a full mask (all values are 1.0) of the
        provided `width` and `height`, also with the
        `number_of_channels` given, and the `dtype`
        provided.
        """
        import numpy as np

        return np.ones((height, width, number_of_channels), dtype = dtype)
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_half_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
    ) -> 'np.ndarray':
        """
        Get a half mask (all values are 0.5) for the
        that fits the dimensions and properties of
        the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        width, height, number_of_channels = _get_width_height_and_number_of_channels_from_input(input)

        return _NumpySelectionMaskGenerator.get_half_mask(
            width = width,
            height = height,
            number_of_channels = number_of_channels,
            dtype = dtype
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_half_mask(
        width: int,
        height: int,
        number_of_channels: int,
        dtype: str = 'f4'
    ) -> 'np.ndarray':
        """
        Get a half mask (all values are 0.5) of the
        provided `width` and `height`, also with the
        `number_of_channels` given, and the `dtype`
        provided.
        """
        return _NumpySelectionMaskGenerator.get_custom_mask(
            width = width,
            height = height,
            number_of_channels = number_of_channels,
            dtype = dtype,
            value = 0.5
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_empty_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
    ) -> 'np.ndarray':
        """
        Get an empty mask (all values are 0.0) for the
        that fits the dimensions and properties of
        the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        width, height, number_of_channels = _get_width_height_and_number_of_channels_from_input(input)

        return _NumpySelectionMaskGenerator.get_empty_mask(
            width = width,
            height = height,
            number_of_channels = number_of_channels,
            dtype = dtype
        )
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_empty_mask(
        width: int,
        height: int,
        number_of_channels: int,
        dtype: str = 'f4'
    ) -> 'np.ndarray':
        """
        Get an empty mask (all values are 0.0) of the
        provided `width` and `height`, also with the
        `number_of_channels` given, and the `dtype`
        provided.
        """
        import numpy as np

        return np.zeros((height, width, number_of_channels), dtype = dtype)
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_random_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
    ) -> 'np.ndarray':
        """
        Get a random mask (all values are random
        values in the range [0.0, 1.0]) that fits
        the dimensions and properties of the `input`
        provided, returning it with the given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        width, height, number_of_channels = _get_width_height_and_number_of_channels_from_input(input)

        return _NumpySelectionMaskGenerator.get_random_mask(
            width = width,
            height = height,
            number_of_channels = number_of_channels,
            dtype = dtype
        )

    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_random_mask(
        width: int,
        height: int,
        number_of_channels: int,
        dtype: str = 'f4'
    ) -> 'np.ndarray':
        """
        Get a full mask (all values are random values
        in the range [0.0, 1.0]) of the provided `width`
        and `height`, also with the `number_of_channels`
        given, and the `dtype` provided.
        """
        import numpy as np

        return np.random.rand(height, width, number_of_channels).astype(dtype)
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_custom_mask_for_input(
        input: Union['moderngl.Texture', 'np.ndarray'],
        dtype: str = 'f4',
        value: float = 0.75
    ) -> 'np.ndarray':
        """
        Get a custom mask (all values are the provided
        `value`, that must be a value in the [0.0, 1.0]
        range) that fits the dimensions and properties
        of the `input` provided, returning it with the
        given `dtype`.

        The `input` can be a numpy array or an opengl
        texture.
        """
        width, height, number_of_channels = _get_width_height_and_number_of_channels_from_input(input)

        return _NumpySelectionMaskGenerator.get_custom_mask(
            width = width,
            height = height,
            number_of_channels = number_of_channels,
            dtype = dtype,
            value = value
        )

    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    def get_custom_mask(
        width: int,
        height: int,
        number_of_channels: int,
        dtype: str = 'f4',
        value: float = 0.75
    ) -> 'np.ndarray':
        """
        Get a custom mask (all values are the provided
        `value`, that must be a value in the [0.0, 1.0]
        range) of the provided `width` and `height`, also
        with the `number_of_channels` given, and the
        `dtype` provided.
        """
        import numpy as np

        return np.full((height, width, number_of_channels), value, dtype = dtype)
    
    @staticmethod
    @requires_dependency('numpy', 'yta_editor_utils', 'numpy')
    @requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
    def _numpy_mask_to_texture(
        mask: 'np.ndarray',
        dtype: str = 'f4',
        opengl_context: Union['moderngl.Context', None] = None
    ) -> 'moderngl.Texture':
        """
        *For internal use only*

        Transform the provided numpy array `mask` into a
        moderngl Texture mask using the context provided
        as `opengl_context` or creating a new one.
        """
        import moderngl
        # TODO: This code is useful to force a mask array
        # with only one channel:
        # opengl_context = OpenGLContext().context
        # mask = mask[..., 0]  # quedarse con un canal
        # h, w = mask.shape[:2]

        # mask_tex = opengl_context.texture((w, h), 1, data = mask.astype('f4').tobytes())

        opengl_context = (
            moderngl.create_context(standalone = True)
            if opengl_context is None else
            opengl_context
        )
        h, w = mask.shape[:2]
        number_of_components = (
            mask.shape[2]
            if mask.ndim == 3 else
            1
        )
        mask_tex = opengl_context.texture(
            (w, h),
            components = number_of_components,
            # f4 is for float32, u1 is for uint8
            data = mask.astype(dtype).tobytes(),
            dtype = dtype
        )
        mask_tex.filter = (moderngl.NEAREST, moderngl.NEAREST)
        # TODO: This below was suggested but it is working
        # without it
        # Make R be all the channels
        #mask_tex.swizzle = 'RRRR'
        #opengl_context.unpack_alignment = 1

        return mask_tex

class SelectionMaskGenerator:
    """
    *Class to be used as a static class*
    
    Class to simplify the way we generate selection
    masks, to be used when handling selection mask
    for CPU or GPU processing.
    """

    texture: _TextureSelectionMaskGenerator = _TextureSelectionMaskGenerator
    """
    Shortcut to the generation of the moderngl textures
    selection masks.
    """
    numpy: _NumpySelectionMaskGenerator = _NumpySelectionMaskGenerator
    """
    Shortcut to the generation of numpy arrays as 
    selection masks.
    """
    
def _get_width_height_and_number_of_channels_from_input(
    input: Union['moderngl.Texture', 'np.ndarray']
) -> tuple[int, int, int]:
    """
    *For internal use only*

    Get the width, the height and the number of
    channels we need to use when creating the
    selection mask to fit the provided `input`.

    They will come as a tuple:
    - `(width, height, number_of_channels)`
    """
    if PythonValidator.is_numpy_array(input):
        return _get_width_height_and_number_of_channels_from_numpy_array(input)
    elif PythonValidator.is_instance_of(input, 'Texture'):
        return _get_width_height_and_number_of_channels_from_moderngl_texture(input)
    else:
        raise Exception('The "input" provided is not a numpy array nor a moderngl Texture.')

@requires_dependency('numpy', 'yta_editor_utils', 'numpy')
def _get_width_height_and_number_of_channels_from_numpy_array(
    input: 'np.ndarray'
) -> tuple[int, int, int]:
    """
    *For internal use only*

    *The "numpy" library is required"

    Get the width, the height and the number of
    channels we need to use when creating the
    selection mask to fit the provided `input`.

    They will come as a tuple:
    - `(width, height, number_of_channels)`
    """
    h, w = input.shape[:2]
    number_of_channels = (
        input.shape[2]
        if input.ndim == 3 else
        1
    )

    return w, h, number_of_channels

@requires_dependency('moderngl', 'yta_editor_utils', 'moderngl')
def _get_width_height_and_number_of_channels_from_moderngl_texture(
    input: 'moderngl.Texture'
) -> tuple[int, int, int]:
    """
    *For internal use only*

    *The "moderngl" library is required"

    Get the width, the height and the number of
    channels we need to use when creating the
    selection mask to fit the provided `input`.

    They will come as a tuple:
    - `(width, height, number_of_channels)`
    """
    w, h = input.size
    number_of_channels = input.components

    return w, h, number_of_channels
